import dpkt.pcapng as dpng
import dpkt

PCAPNG_BT_DSB = 0x0000000A  # Decryption Secret Block


class DecryptionSecretBlock(dpng._PcapngBlock):
    """Decryption Secret block"""

    __hdr__ = (
        ('type', 'I', PCAPNG_BT_DSB),
        ('len', 'I', 12),
        ('secrets_type', 'I', 4),
        ('secrets_length', 'I', 4),
        # (secrets data, variable size),
        # (options, variable size),
        ('_len', 'I', 4)
    )

    def unpack(self, buf):
        dpkt.Packet.unpack(self, buf)
        if self.len > len(buf):
            raise dpkt.NeedData

        # packet data
        po = self.__hdr_len__ - 4  # offset of pkt_data
        self.pkt_data = buf[po:po + self.secrets_length]

        # skip padding between pkt_data and options
        opts_offset = po + dpng._align32b(self.secrets_length)
        self._do_unpack_options(buf, opts_offset)

    def __bytes__(self):
        pkt_buf = self.pkt_data

        pkt_len = len(pkt_buf)
        self.secrets_length = pkt_len
        self.pkt_len = pkt_len

        opts_buf = self._do_pack_options()

        n = self.__hdr_len__ + dpng._align32b(self.secrets_length) + len(opts_buf)
        self.len = n
        self._len = n

        hdr_buf = self._pack_hdr(
            self.type,
            n,
            self.secrets_type,
            self.secrets_length,
            n
        )

        return b''.join([hdr_buf[:-4], dpng._padded(pkt_buf), opts_buf, hdr_buf[-4:]])

    def __len__(self):
        opts_len = sum(len(o) for o in self.opts)
        return self.__hdr_len__ + dpng._align32b(self.caplen) + opts_len


class DecryptionSecretBlockLE(DecryptionSecretBlock):
    __byte_order__ = '<'


class Reader(object):
    """Simple pypcap-compatible pcapng file reader."""

    def __init__(self, fileobj):
        self.name = getattr(fileobj, 'name', '<{0}>'.format(fileobj.__class__.__name__))
        self.__f = fileobj

        shb = dpng.SectionHeaderBlock()
        buf = self.__f.read(shb.__hdr_len__)
        if len(buf) < shb.__hdr_len__:
            raise ValueError('invalid pcapng header')

        # unpack just the header since endianness is not known
        shb.unpack_hdr(buf)
        if shb.type != dpng.PCAPNG_BT_SHB:
            raise ValueError('invalid pcapng header: not a SHB')

        # determine the correct byte order and reload full SHB
        if shb.bom == dpng.BYTE_ORDER_MAGIC_LE:
            self.__le = True
            buf += self.__f.read(dpng._swap32b(shb.len) - shb.__hdr_len__)
            shb = dpng.SectionHeaderBlockLE(buf)
        elif shb.bom == dpng.BYTE_ORDER_MAGIC:
            self.__le = False
            buf += self.__f.read(shb.len - shb.__hdr_len__)
            shb = dpng.SectionHeaderBlock(buf)
        else:
            raise ValueError('unknown endianness')

        # check if this version is supported
        if shb.v_major != dpng.PCAPNG_VERSION_MAJOR:
            raise ValueError('unknown pcapng version {0}.{1}'.format(shb.v_major, shb.v_minor, ))

        # look for a mandatory IDB
        idb = None
        while 1:
            buf = self.__f.read(8)
            if len(buf) < 8:
                break

            blk_type, blk_len = dpng.struct_unpack('<II' if self.__le else '>II', buf)
            buf += self.__f.read(blk_len - 8)

            if blk_type == dpng.PCAPNG_BT_IDB:
                idb = (dpng.InterfaceDescriptionBlockLE(buf) if self.__le
                       else dpng.InterfaceDescriptionBlock(buf))
                break
            # just skip other blocks

        if idb is None:
            raise ValueError('IDB not found')

        # set timestamp resolution and offset
        self._divisor = float(1e6)  # defaults
        self._tsoffset = 0
        for opt in idb.opts:
            if opt.code == dpng.PCAPNG_OPT_IF_TSRESOL:
                # if MSB=0, the remaining bits is a neg power of 10 (e.g. 6 means microsecs)
                # if MSB=1, the remaining bits is a neg power of 2 (e.g. 10 means 1/1024 of second)
                opt_val = dpng.struct_unpack('b', opt.data)[0]
                pow_num = 2 if opt_val & 0b10000000 else 10
                self._divisor = float(pow_num ** (opt_val & 0b01111111))

            elif opt.code == dpng.PCAPNG_OPT_IF_TSOFFSET:
                # 64-bit int that specifies an offset (in seconds) that must be added to the
                # timestamp of each packet
                self._tsoffset = dpng.struct_unpack('<q' if self.__le else '>q', opt.data)[0]

        if idb.linktype in dpng.dltoff:
            self.dloff = dpng.dltoff[idb.linktype]
        else:
            self.dloff = 0

        self.__f.seek(0)

        self.idb = idb
        self.snaplen = idb.snaplen
        self.filter = ''
        self.__iter = iter(self)

    @property
    def fd(self):
        return self.__f.fileno()

    def fileno(self):
        return self.fd

    def datalink(self):
        return self.idb.linktype

    def setfilter(self, value, optimize=1):
        raise NotImplementedError

    def readpkts(self):
        return list(self)

    def __next__(self):
        return next(self.__iter)

    next = __next__  # Python 2 compat

    def dispatch(self, cnt, callback, *args):
        """Collect and process packets with a user callback.

        Return the number of packets processed, or 0 for a savefile.

        Arguments:

        cnt      -- number of packets to process;
                    or 0 to process all packets until EOF
        callback -- function with (timestamp, pkt, *args) prototype
        *args    -- optional arguments passed to callback on execution
        """
        processed = 0
        if cnt > 0:
            for _ in range(cnt):
                try:
                    ts, pkt = next(iter(self))
                except StopIteration:
                    break
                callback(ts, pkt, *args)
                processed += 1
        else:
            for ts, pkt in self:
                callback(ts, pkt, *args)
                processed += 1
        return processed

    def loop(self, callback, *args):
        self.dispatch(0, callback, *args)

    def __iter__(self):
        self.__f.seek(0)
        while 1:
            buf = self.__f.read(8)
            if len(buf) < 8:
                break

            blk_type, blk_len = dpng.struct_unpack('<II' if self.__le else '>II', buf)
            buf += self.__f.read(blk_len - 8)

            if blk_type == dpng.PCAPNG_BT_EPB:
                epb = dpng.EnhancedPacketBlockLE(buf) if self.__le else dpng.EnhancedPacketBlock(buf)
                ts = self._tsoffset + (((epb.ts_high << 32) | epb.ts_low) / self._divisor)
                yield ts, epb.pkt_data
            elif blk_type == dpng.PCAPNG_BT_PB:
                pb = dpng.PacketBlockLE(buf) if self.__le else dpng.PacketBlock(buf)
                ts = self._tsoffset + (((pb.ts_high << 32) | pb.ts_low) / self._divisor)
                yield ts, pb.pkt_data
            elif blk_type == PCAPNG_BT_DSB:
                dsb = DecryptionSecretBlockLE(buf) if self.__le else DecryptionSecretBlock(buf)
                ts = -1
                yield ts, dsb.pkt_data

            # just ignore other blocks



from unittest import TestCase
import tlexport.key_derivator as kd
from cryptography.hazmat.primitives.ciphers.algorithms import ARC4, AES, TripleDES
from cryptography.hazmat.primitives.hashes import SHA256


class KeyGenerationTests(TestCase):
    def setUp(self) -> None:
        pass

    def test_ssl30_master_secret_gen(self):
        pass

    def test_ssl30_key_generation(self):
        self.assertEqual(kd.dev_ssl_30_keys(
            master_secret=bytes.fromhex("2f149d2d9a4c5586edcf4901f98850f6cd90f2daf7b21df3707e2c5c47119bbb45ae175214fd6c48b389a1cfcb0abb12f62736a0366d44aab2a1d430919836ba122e84436ea796c96412f30f8e9af6ca165676b32766235a696d8dec4eb5b6361c98cd37c0894354669d4cd0fe30e1ef"),
            server_random=bytes.fromhex("165676b32766235a696d8dec4eb5b6361c98cd37c0894354669d4cd0fe30e1ef"),
            client_random=bytes.fromhex("f62736a0366d44aab2a1d430919836ba122e84436ea796c96412f30f8e9af6ca"),
            key_length=16,
            mac_length=20,
            key_block_length=72,
            cipher_algo=ARC4,
            use_aead=0),

        {
        "client_write_MAC_secret": bytes.fromhex("27b0f02df5e0eac4f1c2f169723f489dc59dc72a"),
        "server_write_MAC_secret": bytes.fromhex("08d807876acd708ac8be4aa736e95c811c5ee834"),
        "client_write_key": bytes.fromhex("4317dade5961a2c2ec2a159e378e3fbf"),
        "server_write_key": bytes.fromhex("094248fe2aa2550eb92b9adefd024b8c"),
        "client_write_IV": bytes.fromhex("ef9fa205"),
        "server_write_IV": bytes.fromhex("f1725837")
        })

    def test_tls10_master_secret_gen(self):
        pass

    def test_tls10_key_generation(self):
        self.assertEqual(kd.dev_tls_10_11_keys(
            master_secret=bytes.fromhex(
                "de0388430b6540cc6719c520846372b8b07add09d7d4fb22894bc871cbe3679615b96e254928c271796734a99a71b96a"),
            server_random=bytes.fromhex("5ae72e31031a078d035e79a5dfeace288317e1cb33aa8525edfb0642f1b3925d"),
            client_random=bytes.fromhex("ad3a425b23e1fb2c50893f2ca8484baede685205f46238f573d37ae3297faa98"),
            key_length=24,
            mac_length=20,
            key_block_length=88,
            cipher_algo=TripleDES,
            use_aead=0),

            {
                "client_write_MAC_secret": bytes.fromhex("d3c122820fb226ab679e6913e52714aa2293c5c2"),
                "server_write_MAC_secret": bytes.fromhex("1fb0d84eebe0049a69d13d6fcdc6e3f4bb501667"),
                "client_write_key": bytes.fromhex("2d7afb026c41e24f987945b79c317b3b9adbc7e243fa5f51"),
                "server_write_key": bytes.fromhex("57d65d2c1c58d69270ff1d8af3374c0917ee5e618fb303f6"),
                "client_write_IV": bytes.fromhex("a688a99db4fa773d"),
                "server_write_IV": bytes.fromhex("006d9f3e58e97944")
            })

    def test_tls11_master_secret_gen(self):
        pass

    def test_tls11_key_generation(self):
        pass

    def test_tls12_master_secret_gen(self):
        self.assertEqual(kd.gen_master_secret_tls_12(
            pm_secret=bytes.fromhex("df4a291baa1eb7cfa6934b29b474baad2697e29f1f920dcc77c8a0a088447624"),
            client_random=bytes.fromhex("000102030405060708090a0b0c0d0e0f101112131415161718191a1b1c1d1e1f"),
            server_random=bytes.fromhex("707172737475767778797a7b7c7d7e7f808182838485868788898a8b8c8d8e8f")),

            bytes.fromhex(
                "916abf9da55973e13614ae0a3f5d3f37b023ba129aee02cc9134338127cd7049781c8e19fc1eb2a7387ac06ae237344c"))

    def test_tls12_key_generation(self):
        self.assertEqual(kd.dev_tls_12_keys(
            master_secret=bytes.fromhex("916abf9da55973e13614ae0a3f5d3f37b023ba129aee02cc9134338127cd7049781c8e19fc1eb2a7387ac06ae237344c"),
            client_random=bytes.fromhex("000102030405060708090a0b0c0d0e0f101112131415161718191a1b1c1d1e1f"),
            server_random=bytes.fromhex("707172737475767778797a7b7c7d7e7f808182838485868788898a8b8c8d8e8f"),
            key_length=16,
            mac_length=20,
            key_block_length=104,
            cipher_algo=AES,
            use_aead=0,
            mac_function=SHA256),

            {
                "client_write_MAC_secret": bytes.fromhex("1b7d117c7d5f690bc263cae8ef60af0f1878acc2"),
                "server_write_MAC_secret":  bytes.fromhex("2ad8bdd8c601a617126f63540eb20906f781fad2"),
                "client_write_key": bytes.fromhex("f656d037b173ef3e11169f27231a84b6"),
                "server_write_key": bytes.fromhex("752a18e7a9fcb7cbcdd8f98dd8f769eb"),
                "client_write_IV":  bytes.fromhex("a0d2550c9238eebfef5c32251abb67d6"),
                "server_write_IV": bytes.fromhex("434528db4937d540d393135e06a11bb8"),
            })

    def test_tls13_key_generation(self):
        self.assertEqual(kd.dev_tls_10_11_keys(
            master_secret=bytes.fromhex(
                "de0388430b6540cc6719c520846372b8b07add09d7d4fb22894bc871cbe3679615b96e254928c271796734a99a71b96a"),
            server_random=bytes.fromhex("5ae72e31031a078d035e79a5dfeace288317e1cb33aa8525edfb0642f1b3925d"),
            client_random=bytes.fromhex("ad3a425b23e1fb2c50893f2ca8484baede685205f46238f573d37ae3297faa98"),
            key_length=24,
            mac_length=20,
            key_block_length=88,
            cipher_algo=TripleDES,
            use_aead=0),

            {
                "client_write_MAC_secret": bytes.fromhex("d3c122820fb226ab679e6913e52714aa2293c5c2"),
                "server_write_MAC_secret": bytes.fromhex("1fb0d84eebe0049a69d13d6fcdc6e3f4bb501667"),
                "client_write_key": bytes.fromhex("2d7afb026c41e24f987945b79c317b3b9adbc7e243fa5f51"),
                "server_write_key": bytes.fromhex("57d65d2c1c58d69270ff1d8af3374c0917ee5e618fb303f6"),
                "client_write_IV": bytes.fromhex("a688a99db4fa773d"),
                "server_write_IV": bytes.fromhex("006d9f3e58e97944")
            })



from tlexport.packet import Packet


# data class for better handling of tls records
class TlsRecord:
    def __init__(self, binary, metadata: list[Packet], isserver) -> None:
        self.binary = binary[5:]
        self.record_type = binary[0]
        self.record_version = binary[1:3]
        self.record_length = binary[3:5]
        self.metadata = metadata
        self.isserver = isserver

        self.raw = binary



from enum import Enum


class KeyType(Enum):
    RSA = 1
    CLIENT_RANDOM = 2
    CLIENT_EARLY_TRAFFIC_SECRET = 3
    CLIENT_HANDSHAKE_TRAFFIC_SECRET = 4
    SERVER_HANDSHAKE_TRAFFIC_SECRET = 5
    CLIENT_TRAFFIC_SECRET_0 = 6
    SERVER_TRAFFIC_SECRET_0 = 7
    EARLY_EXPORTER_SECRET = 8
    EXPORTER_SECRET = 9
    NONE = 0


class Key:
    def __init__(self, sec_string: str) -> None:
        [ktype, cl_rand, sec] = sec_string.split(" ")
        self.key_type = self.get_key_type(ktype)
        self.client_random = bytes.fromhex(cl_rand)
        self.secret = bytes.fromhex(sec)

    def get_key_type(self, ktype):
        if ktype == "RSA":
            return KeyType.RSA
        elif ktype == "CLIENT_RANDOM":
            return KeyType.CLIENT_RANDOM
        elif ktype == "CLIENT_EARLY_TRAFFIC_SECRET":
            return KeyType.CLIENT_EARLY_TRAFFIC_SECRET
        elif ktype == "CLIENT_HANDSHAKE_TRAFFIC_SECRET":
            return KeyType.CLIENT_HANDSHAKE_TRAFFIC_SECRET
        elif ktype == "SERVER_HANDSHAKE_TRAFFIC_SECRET":
            return KeyType.SERVER_HANDSHAKE_TRAFFIC_SECRET
        elif ktype == "CLIENT_TRAFFIC_SECRET_0":
            return KeyType.CLIENT_TRAFFIC_SECRET_0
        elif ktype == "SERVER_TRAFFIC_SECRET_0":
            return KeyType.SERVER_TRAFFIC_SECRET_0
        elif ktype == "EARLY_EXPORTER_SECRET":
            return KeyType.EARLY_EXPORTER_SECRET
        elif ktype == "EXPORTER_SECRET":
            return KeyType.EXPORTER_SECRET
        else:
            return KeyType.NONE



import logging

class CustomFormatter(logging.Formatter):
    """Custom logging formatter to add colors and custom prefixes, with dynamic format based on log level."""
    
    # Define prefix, color, and format for each log level
    FORMAT = {
        logging.DEBUG: "\033[96m[DEBUG] %(filename)s : %(message)s\033[0m",  # Cyan for DEBUG, including file name
        logging.INFO: "\033[92m[+] %(message)s\033[0m",  # Green for INFO
        logging.WARNING: "\033[93m[W] %(message)s\033[0m",  # Orange for WARNING
        logging.ERROR: "\033[91m[-] %(message)s\033[0m",  # Red for ERROR
    }
    
    def format(self, record):
        self._style._fmt = self.FORMAT.get(record.levelno, self.FORMAT[logging.ERROR])  # Default to ERROR format
        return logging.Formatter.format(self, record)


class LogFilter(logging.Filter):
    def __init__(self, files_to_filter):
        self.files = files_to_filter

    def filter(self, record: logging.LogRecord) -> bool:
        if record.filename in self.files:
            return True
        return False


def set_logger(args):
    log_level = logging.ERROR  # Default to ERROR
    match args.debug:
        case "INFO":
            log_level = logging.INFO
        case "WARNING":
            log_level = logging.WARNING
        case "DEBUG":
            log_level = logging.DEBUG
        case _:
            log_level = logging.ERROR

    log_filter = logging.Filter()

    if args.filter is not None:
        log_filter = LogFilter(args.filter)
    logging.basicConfig(level=log_level)

    logger = logging.getLogger()

    logger.addFilter(log_filter)

    # Remove all handlers associated with the logger object.
    for handler in logger.handlers[:]:
        logger.removeHandler(handler)

    # Create a console handler with the custom formatter
    console_handler = logging.StreamHandler()
    console_handler.setFormatter(CustomFormatter())
    logger.addHandler(console_handler)



import logging
import math

from cryptography.hazmat.primitives import hashes, hmac
from cryptography.hazmat.primitives.ciphers import algorithms
from cryptography.hazmat.primitives.ciphers.aead import ChaCha20Poly1305
from cryptography.hazmat.primitives.kdf.hkdf import HKDFExpand


def dev_ssl_30_keys(master_secret, server_random, client_random, key_length, mac_length, key_block_length, cipher_algo,
                    use_aead):
    iv_length = 4
    if cipher_algo in [algorithms.AES, algorithms.Camellia]:
        iv_length = 16
    elif cipher_algo == algorithms.TripleDES:
        iv_length = 8

    if use_aead:
        mac_length = 0

    key_block = prf_ssl_30(master_secret, client_random, server_random, key_block_length + 2 * iv_length, 0)

    keys = {
        "client_write_MAC_secret": key_block[0: mac_length],
        "server_write_MAC_secret": key_block[mac_length: 2 * mac_length],
        "client_write_key": key_block[2 * mac_length: 2 * mac_length + key_length],
        "server_write_key": key_block[2 * mac_length + key_length: 2 * mac_length + 2 * key_length],
        "client_write_IV": key_block[2 * mac_length + 2 * key_length: 2 * mac_length + 2 * key_length + iv_length],
        "server_write_IV": key_block[
                           2 * mac_length + 2 * key_length + iv_length: 2 * mac_length + 2 * key_length + 2 * iv_length]
    }

    logging_string = "Session Keys: "
    for k in keys:
        logging_string += f"{k}: {keys[k].hex()}\n"

    logging.info(f"{logging_string}")

    return keys


def dev_tls_10_11_keys(master_secret, server_random, client_random, key_length, mac_length, key_block_length,
                       cipher_algo, use_aead):
    iv_length = 4
    if cipher_algo in [algorithms.AES, algorithms.Camellia]:
        iv_length = 16
    elif cipher_algo == algorithms.TripleDES:
        iv_length = 8

    if use_aead:
        mac_length = 0

    key_block = prf_tls_10_11(master_secret, client_random, server_random, b"key expansion",
                              key_block_length + 2 * iv_length, 0)

    keys = {
        "client_write_MAC_secret": key_block[0: mac_length],
        "server_write_MAC_secret": key_block[mac_length: mac_length * 2],
        "client_write_key": key_block[mac_length * 2: mac_length * 2 + key_length],
        "server_write_key": key_block[mac_length * 2 + key_length: mac_length * 2 + key_length * 2],
        "client_write_IV": key_block[mac_length * 2 + key_length * 2: mac_length * 2 + key_length * 2 + iv_length],
        "server_write_IV": key_block[
                           mac_length * 2 + key_length * 2 + iv_length: mac_length * 2 + key_length * 2 + iv_length * 2]

    }

    logging_string = ""
    for k in keys:
        logging_string += f"{k}: {keys[k].hex()}\n"

    logging.info(f"{logging_string}")

    return keys


def dev_tls_12_keys(master_secret, client_random, server_random, key_length, mac_length, key_block_length, cipher_algo,
                    use_aead, mac_function):
    iv_length = 4
    if cipher_algo == ChaCha20Poly1305:
        iv_length = 12

    if cipher_algo in [algorithms.AES, algorithms.Camellia]:
        iv_length = 16

    if use_aead:
        mac_length = 0

    key_block = prf_tls_12(master_secret, client_random, server_random, b'key expansion',
                           key_block_length + 2 * iv_length, mac_function)

    keys = {
        "client_write_MAC_secret": key_block[0: mac_length],
        "server_write_MAC_secret": key_block[mac_length: mac_length * 2],
        "client_write_key": key_block[mac_length * 2: mac_length * 2 + key_length],
        "server_write_key": key_block[mac_length * 2 + key_length: mac_length * 2 + key_length * 2],
        "client_write_IV": key_block[mac_length * 2 + key_length * 2: mac_length * 2 + key_length * 2 + iv_length],
        "server_write_IV": key_block[
                           mac_length * 2 + key_length * 2 + iv_length: mac_length * 2 + key_length * 2 + 2 * iv_length]

    }

    logging_string = ""
    for k in keys:
        logging_string += f"{k}: {keys[k].hex()}\n"

    logging.info(f"{logging_string}")

    return keys


def dev_tls_13_keys(secret_list, key_length, hash_fun: hashes.HashAlgorithm):
    # key length dependent on algorithm iv size always 12
    key_length = int(key_length).to_bytes(2, 'big')
    iv_length = b'\x00\x0c'

    key_label = b'tls13 key'
    iv_label = b'tls13 iv'

    iv_label_len = b'\x08'
    key_label_len = b'\x09'

    iv_info = iv_length + iv_label_len + iv_label + b'\x00'
    key_info = key_length + key_label_len + key_label + b'\x00'

    # In case handshake secrets are not available
    client_handshake_key = None
    client_handshake_iv = None
    server_handshake_key = None
    server_handshake_iv = None

    for secret in secret_list:
        if secret.label == "CLIENT_HANDSHAKE_TRAFFIC_SECRET":
            client_handshake_key = HKDFExpand(hash_fun, int.from_bytes(key_length, 'big'), key_info).derive(
                bytes.fromhex(secret.value))
            client_handshake_iv = HKDFExpand(hash_fun, 12, iv_info).derive(bytes.fromhex(secret.value))

        elif secret.label == "SERVER_HANDSHAKE_TRAFFIC_SECRET":
            server_handshake_key = HKDFExpand(hash_fun, int.from_bytes(key_length, 'big'), key_info).derive(
                bytes.fromhex(secret.value))
            server_handshake_iv = HKDFExpand(hash_fun, 12, iv_info).derive(bytes.fromhex(secret.value))

        elif secret.label == "CLIENT_TRAFFIC_SECRET_0":
            client_application_key = HKDFExpand(hash_fun, int.from_bytes(key_length, 'big'), key_info).derive(
                bytes.fromhex(secret.value))
            client_application_iv = HKDFExpand(hash_fun, 12, iv_info).derive(bytes.fromhex(secret.value))

        elif secret.label == "SERVER_TRAFFIC_SECRET_0":
            server_application_key = HKDFExpand(hash_fun, int.from_bytes(key_length, 'big'), key_info).derive(
                bytes.fromhex(secret.value))
            server_application_iv = HKDFExpand(hash_fun, 12, iv_info).derive(bytes.fromhex(secret.value))

    keys = {
        "client_handshake_traffic_secret": client_handshake_key,
        "server_handshake_traffic_secret": server_handshake_key,
        "client_application_traffic_secret_0": client_application_key,
        "server_application_traffic_secret_0": server_application_key,
        "client_handshake_iv": client_handshake_iv,
        "server_handshake_iv": server_handshake_iv,
        "client_application_iv": client_application_iv,
        "server_application_iv": server_application_iv
    }

    logging_string = ""
    for k in keys:
        logging_string += f"{k}: {keys[k].hex()}\n"

    logging.info(f"{logging_string}")

    return keys


def gen_master_secret_ssl_30(pm_secret, client_random, server_random):
    master_secret = prf_ssl_30(pm_secret, client_random, server_random, 48, 1)
    logging.info(f"Master Secret: {master_secret}")
    return master_secret


def gen_master_secret_tls_10_11(pm_secret, client_random, server_random):
    master_secret = prf_tls_10_11(pm_secret, client_random, server_random, b'master secret', 48, 1)
    logging.info(f"Master Secret: {master_secret}")
    return master_secret


def gen_master_secret_tls_12(pm_secret, client_random, server_random):
    seed = b'master secret' + client_random + server_random

    a0 = seed

    h = hmac.HMAC(pm_secret, hashes.SHA256())
    h.update(a0)
    a1 = h.finalize()
    h = hmac.HMAC(pm_secret, hashes.SHA256())
    h.update(a1)
    a2 = h.finalize()

    h = hmac.HMAC(pm_secret, hashes.SHA256())
    h.update(a1 + seed)
    p1 = h.finalize()

    h = hmac.HMAC(pm_secret, hashes.SHA256())
    h.update(a2 + seed)
    p2 = h.finalize()

    master_secret = p1 + p2[:16]

    logging.info(f"Master Secret: {master_secret}")

    return master_secret


def prf_ssl_30(secret, client_random, server_random, length, non_key):
    key_block = b''
    sec_bits = 'ABCDEFGHIJ'
    counter = 1

    while len(key_block) < length:
        md5 = hashes.Hash(hashes.MD5())
        sha1 = hashes.Hash(hashes.SHA1())
        if non_key:
            sha1.update(bytes(counter * sec_bits[counter - 1], 'utf-8') + secret + client_random + server_random)
        else:
            sha1.update(bytes(counter * sec_bits[counter - 1], 'utf-8') + secret + server_random + client_random)

        a = sha1.finalize()
        md5.update(secret + a)
        key_block = key_block + md5.finalize()

        counter += 1

    return key_block[:length]


def prf_tls_10_11(secret, client_random, server_random, label, length, non_key):
    if non_key:
        seed = label + client_random + server_random
    else:
        seed = label + server_random + client_random

    l_s = len(secret)
    l_s1 = l_s2 = math.ceil(l_s / 2)
    s1 = secret[:l_s1]
    s2 = secret[l_s2:]

    a0 = seed

    p_md5 = b''
    md_5_count = 0

    while len(p_md5) < length:
        md_5_count += 1
        h1 = hmac.HMAC(s1, hashes.MD5())
        h1.update(a0)
        a1 = h1.finalize()

        h = hmac.HMAC(s1, hashes.MD5())
        h.update(a1 + seed)
        p_md5 = p_md5 + h.finalize()
        a0 = a1

    a0 = seed
    p_sha1 = b''
    sha_1_count = 0
    while len(p_sha1) < length:
        sha_1_count += 1
        h1 = hmac.HMAC(s2, hashes.SHA1())
        h1.update(a0)
        a1 = h1.finalize()

        h = hmac.HMAC(s2, hashes.SHA1())
        h.update(a1 + seed)
        p_sha1 = p_sha1 + h.finalize()
        a0 = a1

    secret_block = bytearray([b1 ^ b2 for b1, b2 in zip(p_md5, p_sha1)])

    return secret_block[:length]


def prf_tls_12(secret, client_random, server_random, label, length, mac_function):
    mac = hashes.SHA256

    if mac_function == hashes.SHA384:
        mac = hashes.SHA384

    seed = label + server_random + client_random

    a0 = seed
    secret_block = b''

    while len(secret_block) < length:
        h = hmac.HMAC(secret, mac())
        h.update(a0)

        a1 = h.finalize()

        h = hmac.HMAC(secret, mac())
        h.update(a1 + seed)
        secret_block = secret_block + h.finalize()
        a0 = a1

    return secret_block[:length]



import copy
import logging

from tlexport.packet import Packet


def ones_complement_checksum(byte_arr: bytearray) -> bytearray:
    checksum_arr = copy.deepcopy(byte_arr)

    # pad to multiples of 16 Bit
    if len(checksum_arr) % 2 != 0:
        checksum_arr.extend(b'\x00')

    checksum = 0
    # Sum up all 16-Bit Chunks
    for i in range(0, len(checksum_arr), 2):
        checksum += int.from_bytes(checksum_arr[i:i + 2], "big")

    # short checksum to 16 Bit
    while checksum > 65536:
        first = checksum >> 16
        last = checksum & 0xFFFF
        checksum = first + last

    # ones complement of checksum
    out_arr = bytearray(checksum.to_bytes(2, byteorder='big'))
    for i in range(2):
        out_arr[i] = ~out_arr[i] + 256

    return out_arr


def calculate_checksum_tcp(packet: Packet):
    logging.info("")
    logging.info("TCP Checksum")
    pseudo_header = bytearray(b'')

    # IPv4
    if not packet.ipv6_packet:
        pseudo_header.extend(packet.ip_src)
        pseudo_header.extend(packet.ip_dst)
        pseudo_header.extend(b'\x00')
        pseudo_header.extend(packet.ip.p.to_bytes(1, 'big'))
        pseudo_header.extend(len(packet.tcp).to_bytes(2, 'big'))

    # IPv6
    if packet.ipv6_packet:
        pseudo_header.extend(packet.ip_src)
        pseudo_header.extend(packet.ip_dst)
        pseudo_header.extend(len(packet.tcp).to_bytes(4, 'big'))
        pseudo_header.extend(b'\x00\x00\x00')
        pseudo_header.extend(packet.ip.nxt.to_bytes(1, 'big'))

    # TCP Body
    tcp_data = bytearray(bytes(packet.tcp))
    tcp_data[16:18] = bytearray(b'\x00\x00')
    logging.info(f"pseudo header: 0x{pseudo_header.hex()}")
    logging.info(f"tcp data: 0x{tcp_data.hex()}")

    pseudo_header.extend(tcp_data)

    calculated_checksum = ones_complement_checksum(pseudo_header)
    packet_checksum = packet.tcp.sum.to_bytes(2, 'big')

    logging.info(f"expected checksum: 0x{calculated_checksum.hex()}, packet checksum: 0x{packet_checksum.hex()}")

    return calculated_checksum == packet_checksum



from unittest import TestCase
from tlexport.cipher_suite_parser import cipher_suites, split_cipher_suite
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.ciphers import algorithms, modes
from cryptography.hazmat.primitives.ciphers import aead

import sys

sys.path.append("../tlexport")


class TestCipherSuiteParser(TestCase):
    def test_cipher_suites(self):
        for suite in cipher_suites:
            match suite:
                case b'\x13\x02':
                    self.assertEqual(split_cipher_suite(suite),
                                     {"CryptoAlgo": (aead.AESGCM, 1),
                                      "Mode": (modes.GCM, 1),
                                      "MAC": hashes.SHA384,
                                      "KeyLength": 32,
                                      "TagLength": 16})

                case b'\x13\x03':
                    self.assertEqual(split_cipher_suite(suite),
                                     {"CryptoAlgo": (aead.ChaCha20Poly1305, 1),
                                      "Mode": (aead.ChaCha20Poly1305, 1),
                                      "MAC": hashes.SHA256,
                                      "KeyLength": 32,
                                      "TagLength": 16})
                case b'\x13\x01':
                    self.assertEqual(split_cipher_suite(suite),
                                     {"CryptoAlgo": (aead.AESGCM, 1),
                                      "Mode": (modes.GCM, 1),
                                      "MAC": hashes.SHA256,
                                      "KeyLength": 16,
                                      "TagLength": 16})
                case b'\x13\x04':
                    self.assertEqual(split_cipher_suite(suite),
                                     {"CryptoAlgo": (aead.AESCCM, 1),
                                      "Mode": (aead.AESCCM, 1),
                                      "MAC": hashes.SHA256,
                                      "KeyLength": 16,
                                      "TagLength": 16})
                case b'\x13\x05':
                    self.assertEqual(split_cipher_suite(suite),
                                     {"CryptoAlgo": (aead.AESCCM, 1),
                                      "Mode": (aead.AESCCM, 1),
                                      "MAC": hashes.SHA256,
                                      "KeyLength": 16,
                                      "TagLength": 8})
                case b'\xC0\x83':
                    self.assertEqual(split_cipher_suite(suite),
                                     {"CryptoAlgo": (algorithms.Camellia, 0),
                                      "Mode": (modes.GCM, 1),
                                      "MAC": hashes.SHA384,
                                      "KeyLength": 32,
                                      "TagLength": 16})
                case b'\x00\x24':
                    self.assertEqual(split_cipher_suite(suite),
                                     {"CryptoAlgo": (algorithms.ARC4, 0),
                                      "Mode": (None, 0),
                                      "MAC": hashes.MD5,
                                      "KeyLength": 16,
                                      "TagLength": 16})
                case b'\xC0\x08':
                    self.assertEqual(split_cipher_suite(suite),
                                     {"CryptoAlgo": (algorithms.TripleDES, 0),
                                      "Mode": (modes.CBC, 0),
                                      "MAC": hashes.SHA1,
                                      "KeyLength": 24,
                                      "TagLength": 16})




import logging
import warnings

from tlexport.packet import Packet
from tlexport.tlsrecord import TlsRecord
from tlexport.tlsversion import TlsVersion
import tlexport.key_derivator as key_derivator
import tlexport.cipher_suite_parser as cipher_suite_parser
from tlexport.decryptor import Decryptor
from tlexport.output_builder import OutputBuilder

from ipaddress import IPv6Address, IPv4Address

# Suppress the deprecation warning from the cryptography module.
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from cryptography.hazmat.primitives.ciphers.algorithms import AES, TripleDES, IDEA, Camellia
    from cryptography.hazmat.primitives.ciphers.aead import AESCCM, AESGCM


class Session:
    def __init__(self, packet: Packet, server_ports, keylog, portmap) -> None:
        self.start_packet = packet
        self.keylog = keylog

        self.set_client_and_server_ports(packet, server_ports)

        self.server_tls_records = []
        self.client_tls_records = []
        self.server_packet_buffer = []
        self.client_packet_buffer = []
        self.packet_buffer = []

        self.server_counter = 0
        self.client_counter = 0

        self.seen_packets_server = []
        self.seen_packets_client = []

        self.can_decrypt = False
        self.client_hello_seen = False

        self.server_cipher_change = False
        self.client_cipher_change = False

        self.decryptor: Decryptor
        self.decryptor = None

        self.handle_packet(packet)

        self.application_traffic = []

        self.portmap = portmap

    # search SSLKEYLOG for session log data
    def find_session_secrets(self):
        is_handshake_secret = 0  # Only for TLS 1.3
        secrets = []
        for secret in self.keylog:
            if secret.client_random.lower() == self.client_random.hex().lower():
                if secret.label == "CLIENT_HANDSHAKE_TRAFFIC_SECRET":
                    is_handshake_secret += 1
                elif secret.label == "SERVER_HANDSHAKE_TRAFFIC_SECRET":
                    is_handshake_secret += 2
                secrets.append(secret)

        if is_handshake_secret < 2 and self.tls_version == TlsVersion.TLS13:
            logging.warning(f"Missing CLIENT_HANDSHAKE_TRAFFIC_SECRET")
        elif is_handshake_secret < 3 and self.tls_version == TlsVersion.TLS13:
            logging.warning(f"Missing SERVER_HANDSHAKE_TRAFFIC_SECRET")
        elif is_handshake_secret == 3 and self.tls_version == TlsVersion.TLS13:
            logging.warning(f"Missing CLIENT_HANDSHAKE_TRAFFIC_SECRET and SERVER_HANDSHAKE_TRAFFIC_SECRET")

        logging_string = "SSLKEYLOGFILE Content: "
        for secret in secrets:
            logging_string += f"{secret.label}: {secret.value}"
        logging.info(logging_string)

        return secrets

    # generate session keys from SSLKEYLOG
    def generate_keys(self, tls_version, server_cipher_suite, client_random, server_random):
        cipher_suite = cipher_suite_parser.split_cipher_suite(bytes(server_cipher_suite))

        if cipher_suite is None:
            self.can_decrypt = False
            return

        key_length = cipher_suite["KeyLength"]
        mac_length = cipher_suite["MAC"].digest_size

        secret_list = self.find_session_secrets()

        if len(secret_list) == 0:
            logging.error(f"Missing Secrets\n"
                          f"Server IP: {self.binary_to_ip(self.server_ip)}\n"
                          f"Server Port: {self.server_port}\n"
                          f"Client IP: {self.binary_to_ip(self.client_ip)}\n"
                          f"Client Port: {self.client_port}")
            self.can_decrypt = False
            return

        try:
            secret = secret_list[0]
        except IndexError:
            logging.error(f"Missing Secrets\n"
                          f"Server IP: {self.binary_to_ip(self.server_ip)}\n"
                          f"Server Port: {self.server_port}\n"
                          f"Client IP: {self.binary_to_ip(self.client_ip)}\n"
                          f"Client Port: {self.client_port}")
            self.can_decrypt = False
            return

        match tls_version:
            case TlsVersion.TLS13:
                keys = key_derivator.dev_tls_13_keys(secret_list, key_length, cipher_suite["MAC"]())

            case TlsVersion.TLS12:
                if secret.label == "CLIENT_RANDOM":
                    keys = key_derivator.dev_tls_12_keys(bytes.fromhex(secret.value), client_random, server_random,
                                                         key_length, mac_length, 2 * key_length + 2 * mac_length,
                                                         cipher_suite["CryptoAlgo"][0],
                                                         cipher_suite["Mode"][1], cipher_suite["MAC"])
                elif secret.label == "RSA":
                    master_secret = key_derivator.gen_master_secret_tls_12(bytes.fromhex(secret.value), client_random,
                                                                           server_random)
                    keys = key_derivator.dev_tls_12_keys(master_secret, client_random, server_random, key_length,
                                                         mac_length, 2 * key_length + 2 * mac_length,
                                                         cipher_suite["CryptoAlgo"][0],
                                                         cipher_suite["Mode"][1], cipher_suite["MAC"])

            case TlsVersion.TLS10 | TlsVersion.TLS11:
                if secret.label == "CLIENT_RANDOM":
                    keys = key_derivator.dev_tls_10_11_keys(bytes.fromhex(secret.value), server_random, client_random,
                                                            key_length, mac_length, 2 * key_length + 2 * mac_length,
                                                            cipher_suite["CryptoAlgo"][0], cipher_suite["Mode"][1])
                elif secret.label == "RSA":
                    master_secret = key_derivator.gen_master_secret_tls_10_11(bytes.fromhex(secret.value),
                                                                              client_random, server_random)
                    keys = key_derivator.dev_tls_10_11_keys(master_secret, client_random, server_random, key_length,
                                                            mac_length, 2 * key_length + 2 * mac_length,
                                                            cipher_suite["CryptoAlgo"][0], cipher_suite["Mode"][1])

            case TlsVersion.SSL30:
                if secret.label == "CLIENT_RANDOM":
                    keys = key_derivator.dev_ssl_30_keys(bytes.fromhex(secret.value), server_random, client_random,
                                                         key_length, mac_length, 2 * key_length + 2 * mac_length,
                                                         cipher_suite["CryptoAlgo"][0], cipher_suite["CryptoAlgo"][1])
                elif secret.label == "RSA":
                    master_secret = key_derivator.gen_master_secret_ssl_30(bytes.fromhex(secret.value), client_random,
                                                                           server_random)
                    keys = key_derivator.dev_ssl_30_keys(master_secret, client_random, server_random, key_length,
                                                         mac_length, 2 * key_length + 2 * mac_length,
                                                         cipher_suite["CryptoAlgo"][0], cipher_suite["CryptoAlgo"][1])

        # get block size
        block_size = 0
        algo = cipher_suite["CryptoAlgo"][0]
        if algo in [AES, AESCCM, AESGCM, Camellia]:
            block_size = 128
        elif algo in [TripleDES, IDEA]:
            block_size = 64

        self.decryptor = Decryptor(cipher_suite["CryptoAlgo"][0], cipher_suite["Mode"][0], cipher_suite["MAC"], keys,
                                   self.tls_version, cipher_suite["KeyLength"], cipher_suite["MAC"].digest_size,
                                   cipher_suite["TagLength"], block_size, self.extensions, self.compression_method)

    # gets the session metadata from first session packet
    def set_client_and_server_ports(self, packet: Packet, server_ports):
        self.ipv6 = packet.ipv6_packet

        if packet.sport in server_ports:
            self.server_ip = packet.ip_src
            self.server_port = packet.sport
            self.server_mac_addr = packet.ethernet_src

            self.client_ip = packet.ip_dst
            self.client_port = packet.dport
            self.client_mac_addr = packet.ethernet_dst

        else:
            self.server_ip = packet.ip_dst
            self.server_port = packet.dport
            self.server_mac_addr = packet.ethernet_dst

            self.client_ip = packet.ip_src
            self.client_port = packet.sport
            self.client_mac_addr = packet.ethernet_src

    # checks if packet is in session
    def matches_session(self, packet: Packet):
        if (packet.ip_src == self.server_ip and packet.sport == self.server_port
                and packet.ip_dst == self.client_ip and packet.dport == self.client_port):
            return True
        elif (packet.ip_src == self.client_ip and packet.sport == self.client_port
              and packet.ip_dst == self.server_ip and packet.dport == self.server_port):
            return True
        return False

    # add packet to session, if packet is not a duplicate, splitting into server and client packets
    def handle_packet(self, packet: Packet):
        sequence = packet.seq

        if packet.ip_src == self.server_ip and packet.sport == self.server_port:
            if sequence in self.seen_packets_server:
                return

            self.seen_packets_server.append(sequence)

            self.packet_buffer.append(packet)

        else:
            if sequence in self.seen_packets_client:
                return

            self.seen_packets_client.append(sequence)

            self.packet_buffer.append(packet)

    def decrypt(self):
        print(f"[*] Decrypting session: [{self.binary_to_ip(self.server_ip)}:{self.server_port}-{self.binary_to_ip(self.client_ip)}:{self.client_port}]\n")
        logging.info(f"\n---------------------------------------------------------------------\n"
                     f"Decrypting session:\n"
                     f"Server IP: {self.binary_to_ip(self.server_ip)}\n"
                     f"Server Port: {self.server_port}\n"
                     f"Client IP: {self.binary_to_ip(self.client_ip)}\n"
                     f"Client Port: {self.client_port}"
                     f"\n---------------------------------------------------------------------\n")
        self.get_tls_records()
        self.builder = OutputBuilder(self.application_traffic, self.server_ip, self.client_ip, self.server_port,
                                     self.client_port, self.server_mac_addr, self.client_mac_addr, self.portmap)
        return self.builder.build()

    def handle_tls_handshake_record(self, record: TlsRecord, isserver):
        if self.server_cipher_change or self.client_cipher_change:
            try:
                self.handle_handshake_finished(record, isserver)
            except Exception as e:
                logging.warning(f"Could not decrypt Record: Handshake finished")
            return

        match record.binary[0]:
            # client Hello
            case 0x01:
                self.handle_tls_client_hello(record)
            case 0x02:
                self.handle_tls_server_hello(record)
            # ignore the others for now (in TLS 1.3 in application Records)
            case _:
                try:
                    self.handle_handshake_finished(record, isserver)
                except Exception as e:
                    logging.warning(f"Could not decrypt Record: Handshake finished")

    def handle_handshake_finished(self, record, isserver):
        if self.decryptor is None:
            logging.warning(f"Could not decrypt Record: Handshake finished")
            return
        if self.server_cipher_change and isserver and self.can_decrypt:
            _plaintext = self.decryptor.decrypt(record, isserver)

        if self.client_cipher_change and not isserver and self.can_decrypt:
            _plaintext = self.decryptor.decrypt(record, isserver)

    def handle_tls_client_hello(self, record: TlsRecord):
        self.can_decrypt = False
        self.server_cipher_change = False
        self.client_cipher_change = False
        self.client_random = record.binary[6:38]
        logging.info(f"Client Random: {self.client_random.hex()}")
        self.client_hello_seen = True

    def handle_tls_server_hello(self, record: TlsRecord):
        if self.client_hello_seen:
            self.can_decrypt = True

        self.server_random = record.binary[6: 38]
        logging.info(f"Server Random: {self.server_random.hex()}")

        index = 38
        session_id_length = record.binary[index]
        index += session_id_length + 1

        self.ciphersuite = record.binary[index: index + 2]

        self.compression_method = record.binary[index + 2]

        extensions_length = int.from_bytes(record.binary[index + 3: index + 5], 'big')
        extensions_bin = record.binary[index + 5: index + 5 + extensions_length]

        self.extensions = {}

        extensions_index = 0
        while extensions_index < extensions_length:
            extension_length = int.from_bytes(extensions_bin[extensions_index + 2: extensions_index + 4], 'big')
            self.extensions[bytes(extensions_bin[extensions_index:extensions_index + 2])] \
                = extensions_bin[extensions_index + 4: extensions_index + 4 + extension_length]
            extensions_index += extension_length + 4
        is_tls13 = False
        if self.extensions.get(bytes.fromhex("002b")) == bytearray.fromhex("0304"):
            is_tls13 = True

        match int.from_bytes(record.record_version, 'big'):
            case 0x0300:
                self.tls_version = TlsVersion.SSL30
            case 0x0302:
                self.tls_version = TlsVersion.TLS11
            case _:
                if int.from_bytes(record.binary[4:6], 'big') == 0x0301:
                    self.tls_version = TlsVersion.TLS10
                elif int.from_bytes(record.binary[4:6], 'big') == 0x0303:
                    if is_tls13:
                        self.tls_version = TlsVersion.TLS13
                    else:
                        self.tls_version = TlsVersion.TLS12
                else:
                    self.can_decrypt = False
        self.generate_keys(self.tls_version, self.ciphersuite, self.client_random, self.server_random)

    def handle_alert(self, alert_level):
        # Closing Connection as every error leads to immediate termination of connection
        if alert_level == 0x1 and self.tls_version != TlsVersion.TLS13:
            return
        self.can_decrypt = False
        self.client_hello_seen = False

    def handle_decrypted_handshake_record(self, plaintext, isserver):
        index = 0
        while index < len(plaintext):
            handshake_type = plaintext[index]
            length = int.from_bytes(plaintext[index + 1:index + 4], 'big')

            if handshake_type == 20:
                self.decryptor.update_keys(isserver)

            index += length + 4

    def handle_tls_13_application_record(self, record: TlsRecord, isserver):
        try:
            plaintext = self.decryptor.decrypt(record, isserver)
            subrecord_type = plaintext[-1:]
            if subrecord_type == b'\x16':
                self.handle_decrypted_handshake_record(plaintext[:-1], isserver)
                return
            if subrecord_type == b'\x17':
                self.application_traffic.append((plaintext[:-1], record, isserver))
            if subrecord_type == 21:
                self.handle_alert(record.binary[0])
        except Exception as e:
            logging.warning(f"Could not decrypt Record: {self.tls_version} Application Record")

    def handle_tls_application_record(self, record: TlsRecord, isserver):
        try:
            plaintext = self.decryptor.decrypt(record, isserver)

            self.application_traffic.append((plaintext, record, isserver))
        except Exception as e:
            logging.warning(f"Could not decrypt Record: TLS Application Record")

    # consumes and handles a TLS_Record
    def handle_tls_record(self, record: TlsRecord, isserver):
        logging.info("")
        logging.info(f"Record Type: {record.record_type}")
        logging.info(f"Binary: {record.raw.hex()}")
        match record.record_type:
            # Handshake Record
            case 0x16:
                self.handle_tls_handshake_record(record, isserver)
                self.application_traffic.append(None)

            case 0x17:
                if self.decryptor is None:
                    logging.warning(f"Could not decrypt Record: Decryptor is None")
                if self.can_decrypt and self.decryptor is not None:
                    match self.tls_version:
                        case TlsVersion.TLS13:
                            self.handle_tls_13_application_record(record, isserver)
                        case TlsVersion.TLS12 | TlsVersion.TLS11 | TlsVersion.TLS10 | TlsVersion.SSL30:
                            self.handle_tls_application_record(record, isserver)

                        case _:
                            logging.error(f"Unkown TLS Version\n"
                                          f"Server IP: {self.binary_to_ip(self.server_ip)}\n"
                                          f"Server Port: {self.server_port}\n"
                                          f"Client IP: {self.binary_to_ip(self.client_ip)}\n"
                                          f"Client Port: {self.client_port}")
                            self.can_decrypt = False
                            pass

            # Alert Record
            case 0x15:
                self.handle_alert(record.binary[0])
                self.application_traffic.append(None)

            case 0x14:
                if isserver:
                    self.server_cipher_change = True
                else:
                    self.client_cipher_change = True

    def binary_to_ip(self, ip_addr):
        if self.ipv6:
            return IPv6Address(ip_addr)
        else:
            return IPv4Address(ip_addr)

    # extracts TLS_Records from packet buffers
    def get_tls_records(self):
        packet: Packet
        for packet in self.packet_buffer:
            if packet.ip_src == self.server_ip and packet.sport == self.server_port:
                self.server_packet_buffer.append(packet)
                self.extract_server_buf()

                for record in self.server_tls_records:
                    self.handle_tls_record(record, True)

                self.server_tls_records.clear()
            else:
                self.client_packet_buffer.append(packet)
                self.extract_client_buf()

                for record in self.client_tls_records:
                    self.handle_tls_record(record, False)

                self.client_tls_records.clear()

    # extracts packets from session which together contain complete TLS_Records
    def extract_server_buf(self):
        self.server_counter += 1
        self.server_packet_buffer.sort(key=lambda x: x.seq)

        for i in range(0, len(self.server_packet_buffer) - 1):
            if self.server_packet_buffer[i].seq + len(self.server_packet_buffer[i].tls_data) != \
                    self.server_packet_buffer[i + 1].seq:
                # need more packets (missing packets)
                return

        index = 0
        packet_ranges = []
        total_packet_len = 0
        packet_data = bytearray(b'')

        for i in self.server_packet_buffer:
            packet_len = len(i.tls_data)
            packet_ranges.append((total_packet_len, total_packet_len + packet_len, i))
            total_packet_len += packet_len
            packet_data.extend(i.tls_data)

        while True:
            if total_packet_len - index == 0:
                need_data = False
                break
            if total_packet_len - index < 5:
                need_data = True
                break

            record_len = packet_data[index + 3: index + 5]
            record_len = int.from_bytes(record_len, 'big') + 5

            index += record_len

        if not need_data:
            index = 0
            while index != total_packet_len:
                metadata = []
                record_len = packet_data[index + 3: index + 5]
                record_len = int.from_bytes(record_len, 'big') + 5
                for packet_range in packet_ranges:
                    if index < packet_range[1] and index + record_len > packet_range[0]:
                        metadata.append(packet_range[2])

                binary = packet_data[index:index + record_len]

                tls_record = TlsRecord(binary, metadata, True)
                self.server_tls_records.append(tls_record)

                index += record_len
            self.server_packet_buffer.clear()

    # extracts packets from session which together contain complete TLS_Records
    def extract_client_buf(self):
        self.client_counter += 1
        self.client_packet_buffer.sort(key=lambda x: x.seq)

        for i in range(0, len(self.client_packet_buffer) - 1):
            if self.client_packet_buffer[i].seq + len(self.client_packet_buffer[i].tls_data) != \
                    self.client_packet_buffer[i + 1].seq:
                # need more packets (missing packets)
                return

        index = 0
        packet_ranges = []
        total_packet_len = 0
        packet_data = bytearray(b'')

        for i in self.client_packet_buffer:
            packet_len = len(i.tls_data)
            packet_ranges.append((total_packet_len, total_packet_len + packet_len, i))
            total_packet_len += packet_len
            packet_data.extend(i.tls_data)

        while True:
            if total_packet_len - index == 0:
                need_data = False
                break
            if total_packet_len - index < 5:
                need_data = True
                break

            record_len = packet_data[index + 3: index + 5]
            record_len = int.from_bytes(record_len, 'big') + 5

            index += record_len

        if not need_data:
            index = 0
            while index != total_packet_len:
                metadata = []
                record_len = packet_data[index + 3: index + 5]
                record_len = int.from_bytes(record_len, 'big') + 5
                for packet_range in packet_ranges:
                    if index < packet_range[1] and index + record_len > packet_range[0]:
                        metadata.append(packet_range[2])

                binary = packet_data[index:index + record_len]

                tls_record = TlsRecord(binary, metadata, True)
                self.client_tls_records.append(tls_record)

                index += record_len
            self.client_packet_buffer.clear()





from unittest import TestCase
from tlexport.checksums import ones_complement_checksum
from tlexport.checksums import calculate_checksum_tcp
from tlexport.packet import Packet


class TestChecksums(TestCase):
    def setUp(self) -> None:
        self.string_1 = bytearray(
            b'\x7f\x00\x00\x01\x7f\x00\x00\x01\x00\x06\x00('
            b'\x8f\xfe\xad*\xaf;\xe7\xac\x00\x00\x00\x00\xa0\x02\xff\xd7\x00\x00\x00\x00\x02\x04\xff\xd7\x04\x02\x08'
            b'\nX\xf7`\xee\x00\x00\x00\x00\x01\x03\x03\x07')
        self.string_2 = bytearray(
            b'\x7f\x00\x00\x01\x7f\x00\x00\x01\x00\x06\x00('
            b'\xad*\x8f\xfe\xd7\xcfN\xb8\xaf;\xe7\xad\xa0\x12\xff\xcb\x00\x00\x00\x00\x02\x04\xff\xd7\x04\x02\x08\nX'
            b'\xf7`\xeeX\xf7`\xee\x01\x03\x03\x07')

        self.false_string_1 = bytearray(
            b'\x7f\x00\x00\x01\x7f\x00\x00\x01\x00\x06\x00('
            b'\x8f\xfe\xad*\xaf;\xe7\xac\x00\x00\x00\x00\xa0\x02\xff\xd7\x00\x00\x00\x00\x02\x04\xff\xd7')

        self.zeroes = bytearray(b'')

        self.packet_1 = Packet(
            b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x08\x00E\x00\x00<\x11\xdc@\x00@\x06*\xde\x7f\x00\x00'
            b'\x01\x7f\x00\x00\x01\x8f\xfe\xad*\xaf;\xe7\xac\x00\x00\x00\x00\xa0\x02\xff\xd7\xc2\x0a\x00\x00\x02\x04'
            b'\xff\xd7\x04\x02\x08\nX\xf7`\xee\x00\x00\x00\x00\x01\x03\x03\x07',
            1693392019.4963138)
        self.bad_packet_1 = Packet(
            b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x08\x00E\x00\x00<\x00\x00@\x00@\x06<\xba\x7f\x00\x00'
            b'\x01\x7f\x00\x00\x01\xad*\x8f\xfe\xd7\xcfN\xb8\xaf;\xe7\xad\xa0\x12\xff\xcb\xfe0\x00\x00\x02\x04\xff'
            b'\xd7\x04\x02\x08\nX\xf7`\xeeX\xf7`\xee\x01\x03\x03\x07',
            1693392019.4963367)

    def test_ones_complement_checksum(self):
        self.assertEqual(ones_complement_checksum(self.string_1), bytearray.fromhex("c20a"))
        self.assertEqual(ones_complement_checksum(self.string_2), bytearray.fromhex("e197"))
        self.assertNotEqual(ones_complement_checksum(self.false_string_1), bytearray.fromhex("ab12"))

        self.assertEqual(ones_complement_checksum(self.zeroes), bytearray.fromhex("ffff"))

    def test_tcp_checksum(self):
        self.assertEqual(calculate_checksum_tcp(self.packet_1), True)
        self.assertEqual(calculate_checksum_tcp(self.bad_packet_1), False)




import logging
import warnings

# Suppress the deprecation warning from the cryptography module.
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from cryptography.hazmat.primitives import hashes
    from cryptography.hazmat.primitives.ciphers import modes
    from cryptography.hazmat.primitives.ciphers.algorithms import AES, TripleDES, Camellia, IDEA, ARC4, ChaCha20
    from cryptography.hazmat.primitives.ciphers import aead

cipher_suites = {
    # The following 5 cipher suites are the only cipher suites supported by TLSv1.3
    b'\x13\x02': 'TLS_AES_256_GCM_SHA384',
    b'\x13\x03': 'TLS_CHACHA20_POLY1305_SHA256',
    b'\x13\x01': 'TLS_AES_128_GCM_SHA256',
    b'\x13\x04': 'TLS_AES_128_CCM_SHA256',
    b'\x13\x05': 'TLS_AES_128_CCM_8_SHA256',

    # The following cipher suites are other cipher suites recommended by the IANA (Date: September 2023)
    b'\xD0\x05': 'TLS_ECDHE_PSK_WITH_AES_128_CCM_SHA256',
    b'\xD0\x02': 'TLS_ECDHE_PSK_WITH_AES_256_GCM_SHA384',
    b'\xD0\x01': 'TLS_ECDHE_PSK_WITH_AES_128_GCM_SHA256',
    b'\xCC\xAD': 'TLS_DHE_PSK_WITH_CHACHA20_POLY1305_SHA256',
    b'\xCC\xAC': 'TLS_ECDHE_PSK_WITH_CHACHA20_POLY1305_SHA256',
    b'\xCC\xAA': 'TLS_DHE_RSA_WITH_CHACHA20_POLY1305_SHA256',
    b'\xCC\xA9': 'TLS_ECDHE_ECDSA_WITH_CHACHA20_POLY1305_SHA256',
    b'\xCC\xA8': 'TLS_ECDHE_RSA_WITH_CHACHA20_POLY1305_SHA256',
    b'\xC0\xA7': 'TLS_DHE_PSK_WITH_AES_256_CCM',
    b'\xC0\xA6': 'TLS_DHE_PSK_WITH_AES_128_CCM',
    b'\xC0\x9F': 'TLS_DHE_RSA_WITH_AES_256_CCM',
    b'\xC0\x9E': 'TLS_DHE_RSA_WITH_AES_128_CCM',
    b'\xC0\x30': 'TLS_ECDHE_RSA_WITH_AES_256_GCM_SHA384',
    b'\xC0\x2F': 'TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256',
    b'\xC0\x2B': 'TLS_ECDHE_ECDSA_WITH_AES_128_GCM_SHA256',
    b'\xC0\x2C': 'TLS_ECDHE_ECDSA_WITH_AES_256_GCM_SHA384',
    b'\x00\xAA': 'TLS_DHE_PSK_WITH_AES_128_GCM_SHA256',
    b'\x00\xAB': 'TLS_DHE_PSK_WITH_AES_256_GCM_SHA384',
    b'\x00\x9E': 'TLS_DHE_RSA_WITH_AES_128_GCM_SHA256',
    b'\x00\x9F': 'TLS_DHE_RSA_WITH_AES_256_GCM_SHA384',

    # All following cipher suites are cipher suites that are DEPRECATED and NOT RECOMMENDED by the IANA except 'TLS_AES_128_CCM_8_SHA256'

    # Deprecated cipher suites using CAMELLIA
    b'\xC0\x9B': 'TLS_ECDHE_PSK_WITH_CAMELLIA_256_CBC_SHA384',
    b'\xC0\x9A': 'TLS_ECDHE_PSK_WITH_CAMELLIA_128_CBC_SHA256',
    b'\xC0\x99': 'TLS_RSA_PSK_WITH_CAMELLIA_256_CBC_SHA384',
    b'\xC0\x98': 'TLS_RSA_PSK_WITH_CAMELLIA_128_CBC_SHA256',
    b'\xC0\x97': 'TLS_DHE_PSK_WITH_CAMELLIA_256_CBC_SHA384',
    b'\xC0\x96': 'TLS_DHE_PSK_WITH_CAMELLIA_128_CBC_SHA256',
    b'\xC0\x95': 'TLS_PSK_WITH_CAMELLIA_256_CBC_SHA384',
    b'\xC0\x94': 'TLS_PSK_WITH_CAMELLIA_128_CBC_SHA256',

    # b'\xC0\x93': 'TLS_RSA_PSK_WITH_CAMELLIA_256_GCM_SHA384',
    # b'\xC0\x92': 'TLS_RSA_PSK_WITH_CAMELLIA_128_GCM_SHA256',
    # b'\xC0\x91': 'TLS_DHE_PSK_WITH_CAMELLIA_256_GCM_SHA384',
    # b'\xC0\x90': 'TLS_DHE_PSK_WITH_CAMELLIA_128_GCM_SHA256',
    # b'\xC0\x8F': 'TLS_PSK_WITH_CAMELLIA_256_GCM_SHA384',
    # b'\xC0\x8E': 'TLS_PSK_WITH_CAMELLIA_128_GCM_SHA256',
    # b'\xC0\x8D': 'TLS_ECDH_RSA_WITH_CAMELLIA_256_GCM_SHA384',
    # b'\xC0\x8C': 'TLS_ECDH_RSA_WITH_CAMELLIA_128_GCM_SHA256',
    # b'\xC0\x8B': 'TLS_ECDHE_RSA_WITH_CAMELLIA_256_GCM_SHA384',
    # b'\xC0\x8A': 'TLS_ECDHE_RSA_WITH_CAMELLIA_128_GCM_SHA256',
    # b'\xC0\x89': 'TLS_ECDH_ECDSA_WITH_CAMELLIA_256_GCM_SHA384',
    # b'\xC0\x88': 'TLS_ECDH_ECDSA_WITH_CAMELLIA_128_GCM_SHA256',
    # b'\xC0\x87': 'TLS_ECDHE_ECDSA_WITH_CAMELLIA_256_GCM_SHA384',
    # b'\xC0\x86': 'TLS_ECDHE_ECDSA_WITH_CAMELLIA_128_GCM_SHA256',
    # b'\xC0\x85': 'TLS_DH_anon_WITH_CAMELLIA_256_GCM_SHA384',
    # b'\xC0\x84': 'TLS_DH_anon_WITH_CAMELLIA_128_GCM_SHA256',
    # b'\xC0\x83': 'TLS_DH_DSS_WITH_CAMELLIA_256_GCM_SHA384',
    # b'\xC0\x82': 'TLS_DH_DSS_WITH_CAMELLIA_128_GCM_SHA256',
    # b'\xC0\x81': 'TLS_DHE_DSS_WITH_CAMELLIA_256_GCM_SHA384',
    # b'\xC0\x80': 'TLS_DHE_DSS_WITH_CAMELLIA_128_GCM_SHA256',
    # b'\xC0\x7F': 'TLS_DH_RSA_WITH_CAMELLIA_256_GCM_SHA384',
    # b'\xC0\x7E': 'TLS_DH_RSA_WITH_CAMELLIA_128_GCM_SHA256',
    # b'\xC0\x7D': 'TLS_DHE_RSA_WITH_CAMELLIA_256_GCM_SHA384',
    # b'\xC0\x7C': 'TLS_DHE_RSA_WITH_CAMELLIA_128_GCM_SHA256',
    # b'\xC0\x7B': 'TLS_RSA_WITH_CAMELLIA_256_GCM_SHA384',
    # b'\xC0\x7A': 'TLS_RSA_WITH_CAMELLIA_128_GCM_SHA256',

    b'\xC0\x79': 'TLS_ECDH_RSA_WITH_CAMELLIA_256_CBC_SHA384',
    b'\xC0\x78': 'TLS_ECDH_RSA_WITH_CAMELLIA_128_CBC_SHA256',
    b'\xC0\x77': 'TLS_ECDHE_RSA_WITH_CAMELLIA_256_CBC_SHA384',
    b'\xC0\x76': 'TLS_ECDHE_RSA_WITH_CAMELLIA_128_CBC_SHA256',
    b'\xC0\x75': 'TLS_ECDH_ECDSA_WITH_CAMELLIA_256_CBC_SHA384',
    b'\xC0\x74': 'TLS_ECDH_ECDSA_WITH_CAMELLIA_128_CBC_SHA256',
    b'\xC0\x73': 'TLS_ECDHE_ECDSA_WITH_CAMELLIA_256_CBC_SHA384',
    b'\xC0\x72': 'TLS_ECDHE_ECDSA_WITH_CAMELLIA_128_CBC_SHA256',
    b'\x00\xC5': 'TLS_DH_anon_WITH_CAMELLIA_256_CBC_SHA256',
    b'\x00\xC4': 'TLS_DHE_RSA_WITH_CAMELLIA_256_CBC_SHA256',
    b'\x00\xC3': 'TLS_DHE_DSS_WITH_CAMELLIA_256_CBC_SHA256',
    b'\x00\xC2': 'TLS_DH_RSA_WITH_CAMELLIA_256_CBC_SHA256',
    b'\x00\xC1': 'TLS_DH_DSS_WITH_CAMELLIA_256_CBC_SHA256',
    b'\x00\xC0': 'TLS_RSA_WITH_CAMELLIA_256_CBC_SHA256',
    b'\x00\xBF': 'TLS_DH_anon_WITH_CAMELLIA_128_CBC_SHA256',
    b'\x00\xBE': 'TLS_DHE_RSA_WITH_CAMELLIA_128_CBC_SHA256',
    b'\x00\xBD': 'TLS_DHE_DSS_WITH_CAMELLIA_128_CBC_SHA256',
    b'\x00\xBC': 'TLS_DH_RSA_WITH_CAMELLIA_128_CBC_SHA256',
    b'\x00\xBB': 'TLS_DH_DSS_WITH_CAMELLIA_128_CBC_SHA256',
    b'\x00\xBA': 'TLS_RSA_WITH_CAMELLIA_128_CBC_SHA256',
    b'\x00\x89': 'TLS_DH_anon_WITH_CAMELLIA_256_CBC_SHA',
    b'\x00\x88': 'TLS_DHE_RSA_WITH_CAMELLIA_256_CBC_SHA',
    b'\x00\x87': 'TLS_DHE_DSS_WITH_CAMELLIA_256_CBC_SHA',
    b'\x00\x86': 'TLS_DH_RSA_WITH_CAMELLIA_256_CBC_SHA',
    b'\x00\x85': 'TLS_DH_DSS_WITH_CAMELLIA_256_CBC_SHA',
    b'\x00\x84': 'TLS_RSA_WITH_CAMELLIA_256_CBC_SHA',
    b'\x00\x46': 'TLS_DH_anon_WITH_CAMELLIA_128_CBC_SHA',
    b'\x00\x45': 'TLS_DHE_RSA_WITH_CAMELLIA_128_CBC_SHA',
    b'\x00\x44': 'TLS_DHE_DSS_WITH_CAMELLIA_128_CBC_SHA',
    b'\x00\x43': 'TLS_DH_RSA_WITH_CAMELLIA_128_CBC_SHA',
    b'\x00\x42': 'TLS_DH_DSS_WITH_CAMELLIA_128_CBC_SHA',
    b'\x00\x41': 'TLS_RSA_WITH_CAMELLIA_128_CBC_SHA',

    # Deprecated cipher suites using RC4
    b'\xC0\x33': 'TLS_ECDHE_PSK_WITH_RC4_128_SHA',
    b'\xC0\x16': 'TLS_ECDH_anon_WITH_RC4_128_SHA',
    b'\xC0\x11': 'TLS_ECDHE_RSA_WITH_RC4_128_SHA',
    b'\xC0\x0C': 'TLS_ECDH_RSA_WITH_RC4_128_SHA',
    b'\xC0\x07': 'TLS_ECDHE_ECDSA_WITH_RC4_128_SHA',
    b'\xC0\x02': 'TLS_ECDH_ECDSA_WITH_RC4_128_SHA',
    b'\x00\x92': 'TLS_RSA_PSK_WITH_RC4_128_SHA',
    b'\x00\x8E': 'TLS_DHE_PSK_WITH_RC4_128_SHA',
    b'\x00\x8A': 'TLS_PSK_WITH_RC4_128_SHA',
    b'\x00\x24': 'TLS_KRB5_WITH_RC4_128_MD5',
    b'\x00\x20': 'TLS_KRB5_WITH_RC4_128_SHA',
    b'\x00\x18': 'TLS_DH_anon_WITH_RC4_128_MD5',
    b'\x00\x05': 'TLS_RSA_WITH_RC4_128_SHA',
    b'\x00\x04': 'TLS_RSA_WITH_RC4_128_MD5',

    # Deprecated cipher suites using 3DES (Triple-DES)
    b'\xC0\x34': 'TLS_ECDHE_PSK_WITH_3DES_EDE_CBC_SHA',
    b'\xC0\x1C': 'TLS_SRP_SHA_DSS_WITH_3DES_EDE_CBC_SHA',
    b'\xC0\x1B': 'TLS_SRP_SHA_RSA_WITH_3DES_EDE_CBC_SHA',
    b'\xC0\x1A': 'TLS_SRP_SHA_WITH_3DES_EDE_CBC_SHA',
    b'\xC0\x17': 'TLS_ECDH_anon_WITH_3DES_EDE_CBC_SHA',
    b'\xC0\x12': 'TLS_ECDHE_RSA_WITH_3DES_EDE_CBC_SHA',
    b'\xC0\x0D': 'TLS_ECDH_RSA_WITH_3DES_EDE_CBC_SHA',
    b'\xC0\x08': 'TLS_ECDHE_ECDSA_WITH_3DES_EDE_CBC_SHA',
    b'\xC0\x03': 'TLS_ECDH_ECDSA_WITH_3DES_EDE_CBC_SHA',
    b'\x00\x93': 'TLS_RSA_PSK_WITH_3DES_EDE_CBC_SHA',
    b'\x00\x8F': 'TLS_DHE_PSK_WITH_3DES_EDE_CBC_SHA',
    b'\x00\x8B': 'TLS_PSK_WITH_3DES_EDE_CBC_SHA',
    b'\x00\x23': 'TLS_KRB5_WITH_3DES_EDE_CBC_MD5',
    b'\x00\x1F': 'TLS_KRB5_WITH_3DES_EDE_CBC_SHA',
    b'\x00\x1B': 'TLS_DH_anon_WITH_3DES_EDE_CBC_SHA',
    b'\x00\x16': 'TLS_DHE_RSA_WITH_3DES_EDE_CBC_SHA',
    b'\x00\x13': 'TLS_DHE_DSS_WITH_3DES_EDE_CBC_SHA',
    b'\x00\x10': 'TLS_DH_RSA_WITH_3DES_EDE_CBC_SHA',
    b'\x00\x0D': 'TLS_DH_DSS_WITH_3DES_EDE_CBC_SHA',
    b'\x00\x0A': 'TLS_RSA_WITH_3DES_EDE_CBC_SHA',

    # Deprecated cipher suites using ChaCha20
    b'\xCC\xAE': 'TLS_RSA_PSK_WITH_CHACHA20_POLY1305_SHA256',
    b'\xCC\xAB': 'TLS_PSK_WITH_CHACHA20_POLY1305_SHA256',

    # Deprecated cipher suites using AES
    b'\xD0\x03': 'TLS_ECDHE_PSK_WITH_AES_128_CCM_8_SHA256',
    b'\xC0\xB3': 'TLS_ECCPWD_WITH_AES_256_CCM_SHA384',
    b'\xC0\xB2': 'TLS_ECCPWD_WITH_AES_128_CCM_SHA256',
    b'\xC0\xB1': 'TLS_ECCPWD_WITH_AES_256_GCM_SHA384',
    b'\xC0\xB0': 'TLS_ECCPWD_WITH_AES_128_GCM_SHA256',
    b'\xC0\xAF': 'TLS_ECDHE_ECDSA_WITH_AES_256_CCM_8',
    b'\xC0\xAE': 'TLS_ECDHE_ECDSA_WITH_AES_128_CCM_8',
    b'\xC0\xAD': 'TLS_ECDHE_ECDSA_WITH_AES_256_CCM',
    b'\xC0\xAC': 'TLS_ECDHE_ECDSA_WITH_AES_128_CCM',
    b'\xC0\xAB': 'TLS_PSK_DHE_WITH_AES_256_CCM_8',
    b'\xC0\xAA': 'TLS_PSK_DHE_WITH_AES_128_CCM_8',
    b'\xC0\xA9': 'TLS_PSK_WITH_AES_256_CCM_8',
    b'\xC0\xA8': 'TLS_PSK_WITH_AES_128_CCM_8',
    b'\xC0\xA5': 'TLS_PSK_WITH_AES_256_CCM',
    b'\xC0\xA4': 'TLS_PSK_WITH_AES_128_CCM',
    b'\xC0\xA3': 'TLS_DHE_RSA_WITH_AES_256_CCM_8',
    b'\xC0\xA2': 'TLS_DHE_RSA_WITH_AES_128_CCM_8',
    b'\xC0\xA1': 'TLS_RSA_WITH_AES_256_CCM_8',
    b'\xC0\xA0': 'TLS_RSA_WITH_AES_128_CCM_8',
    b'\xC0\x9D': 'TLS_RSA_WITH_AES_256_CCM',
    b'\xC0\x9C': 'TLS_RSA_WITH_AES_128_CCM',
    b'\xC0\x38': 'TLS_ECDHE_PSK_WITH_AES_256_CBC_SHA384',
    b'\xC0\x37': 'TLS_ECDHE_PSK_WITH_AES_128_CBC_SHA256',
    b'\xC0\x36': 'TLS_ECDHE_PSK_WITH_AES_256_CBC_SHA',
    b'\xC0\x35': 'TLS_ECDHE_PSK_WITH_AES_128_CBC_SHA',
    b'\xC0\x32': 'TLS_ECDH_RSA_WITH_AES_256_GCM_SHA384',
    b'\xC0\x31': 'TLS_ECDH_RSA_WITH_AES_128_GCM_SHA256',
    b'\xC0\x2E': 'TLS_ECDH_ECDSA_WITH_AES_256_GCM_SHA384',
    b'\xC0\x2D': 'TLS_ECDH_ECDSA_WITH_AES_128_GCM_SHA256',
    b'\xC0\x2A': 'TLS_ECDH_RSA_WITH_AES_256_CBC_SHA384',
    b'\xC0\x29': 'TLS_ECDH_RSA_WITH_AES_128_CBC_SHA256',
    b'\xC0\x28': 'TLS_ECDHE_RSA_WITH_AES_256_CBC_SHA384',
    b'\xC0\x27': 'TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA256',
    b'\xC0\x26': 'TLS_ECDH_ECDSA_WITH_AES_256_CBC_SHA384',
    b'\xC0\x25': 'TLS_ECDH_ECDSA_WITH_AES_128_CBC_SHA256',
    b'\xC0\x24': 'TLS_ECDHE_ECDSA_WITH_AES_256_CBC_SHA384',
    b'\xC0\x23': 'TLS_ECDHE_ECDSA_WITH_AES_128_CBC_SHA256',
    b'\xC0\x22': 'TLS_SRP_SHA_DSS_WITH_AES_256_CBC_SHA',
    b'\xC0\x21': 'TLS_SRP_SHA_RSA_WITH_AES_256_CBC_SHA',
    b'\xC0\x20': 'TLS_SRP_SHA_WITH_AES_256_CBC_SHA',
    b'\xC0\x1F': 'TLS_SRP_SHA_DSS_WITH_AES_128_CBC_SHA',
    b'\xC0\x1E': 'TLS_SRP_SHA_RSA_WITH_AES_128_CBC_SHA',
    b'\xC0\x1D': 'TLS_SRP_SHA_WITH_AES_128_CBC_SHA',
    b'\xC0\x19': 'TLS_ECDH_anon_WITH_AES_256_CBC_SHA',
    b'\xC0\x18': 'TLS_ECDH_anon_WITH_AES_128_CBC_SHA',
    b'\xC0\x14': 'TLS_ECDHE_RSA_WITH_AES_256_CBC_SHA',
    b'\xC0\x13': 'TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA',
    b'\xC0\x0F': 'TLS_ECDH_RSA_WITH_AES_256_CBC_SHA',
    b'\xC0\x0E': 'TLS_ECDH_RSA_WITH_AES_128_CBC_SHA',
    b'\xC0\x0A': 'TLS_ECDHE_ECDSA_WITH_AES_256_CBC_SHA',
    b'\xC0\x09': 'TLS_ECDHE_ECDSA_WITH_AES_128_CBC_SHA',
    b'\xC0\x05': 'TLS_ECDH_ECDSA_WITH_AES_256_CBC_SHA',
    b'\xC0\x04': 'TLS_ECDH_ECDSA_WITH_AES_128_CBC_SHA',
    b'\x00\xB7': 'TLS_RSA_PSK_WITH_AES_256_CBC_SHA384',
    b'\x00\xB6': 'TLS_RSA_PSK_WITH_AES_128_CBC_SHA256',
    b'\x00\xB3': 'TLS_DHE_PSK_WITH_AES_256_CBC_SHA384',
    b'\x00\xB2': 'TLS_DHE_PSK_WITH_AES_128_CBC_SHA256',
    b'\x00\xAF': 'TLS_PSK_WITH_AES_256_CBC_SHA384',
    b'\x00\xAE': 'TLS_PSK_WITH_AES_128_CBC_SHA256',
    b'\x00\xAD': 'TLS_RSA_PSK_WITH_AES_256_GCM_SHA384',
    b'\x00\xAC': 'TLS_RSA_PSK_WITH_AES_128_GCM_SHA256',
    b'\x00\xA9': 'TLS_PSK_WITH_AES_256_GCM_SHA384',
    b'\x00\xA8': 'TLS_PSK_WITH_AES_128_GCM_SHA256',
    b'\x00\xA7': 'TLS_DH_anon_WITH_AES_256_GCM_SHA384',
    b'\x00\xA6': 'TLS_DH_anon_WITH_AES_128_GCM_SHA256',
    b'\x00\xA5': 'TLS_DH_DSS_WITH_AES_256_GCM_SHA384',
    b'\x00\xA4': 'TLS_DH_DSS_WITH_AES_128_GCM_SHA256',
    b'\x00\xA3': 'TLS_DHE_DSS_WITH_AES_256_GCM_SHA384',
    b'\x00\xA2': 'TLS_DHE_DSS_WITH_AES_128_GCM_SHA256',
    b'\x00\xA1': 'TLS_DH_RSA_WITH_AES_256_GCM_SHA384',
    b'\x00\xA0': 'TLS_DH_RSA_WITH_AES_128_GCM_SHA256',
    b'\x00\x9D': 'TLS_RSA_WITH_AES_256_GCM_SHA384',
    b'\x00\x9C': 'TLS_RSA_WITH_AES_128_GCM_SHA256',
    b'\x00\x95': 'TLS_RSA_PSK_WITH_AES_256_CBC_SHA',
    b'\x00\x94': 'TLS_RSA_PSK_WITH_AES_128_CBC_SHA',
    b'\x00\x91': 'TLS_DHE_PSK_WITH_AES_256_CBC_SHA',
    b'\x00\x90': 'TLS_DHE_PSK_WITH_AES_128_CBC_SHA',
    b'\x00\x8D': 'TLS_PSK_WITH_AES_256_CBC_SHA',
    b'\x00\x8C': 'TLS_PSK_WITH_AES_128_CBC_SHA',
    b'\x00\x6D': 'TLS_DH_anon_WITH_AES_256_CBC_SHA256',
    b'\x00\x6C': 'TLS_DH_anon_WITH_AES_128_CBC_SHA256',
    b'\x00\x6B': 'TLS_DHE_RSA_WITH_AES_256_CBC_SHA256',
    b'\x00\x6A': 'TLS_DHE_DSS_WITH_AES_256_CBC_SHA256',
    b'\x00\x69': 'TLS_DH_RSA_WITH_AES_256_CBC_SHA256',
    b'\x00\x68': 'TLS_DH_DSS_WITH_AES_256_CBC_SHA256',
    b'\x00\x67': 'TLS_DHE_RSA_WITH_AES_128_CBC_SHA256',
    b'\x00\x40': 'TLS_DHE_DSS_WITH_AES_128_CBC_SHA256',
    b'\x00\x3F': 'TLS_DH_RSA_WITH_AES_128_CBC_SHA256',
    b'\x00\x3E': 'TLS_DH_DSS_WITH_AES_128_CBC_SHA256',
    b'\x00\x3D': 'TLS_RSA_WITH_AES_256_CBC_SHA256',
    b'\x00\x3C': 'TLS_RSA_WITH_AES_128_CBC_SHA256',
    b'\x00\x3A': 'TLS_DH_anon_WITH_AES_256_CBC_SHA',
    b'\x00\x39': 'TLS_DHE_RSA_WITH_AES_256_CBC_SHA',
    b'\x00\x38': 'TLS_DHE_DSS_WITH_AES_256_CBC_SHA',
    b'\x00\x37': 'TLS_DH_RSA_WITH_AES_256_CBC_SHA',
    b'\x00\x36': 'TLS_DH_DSS_WITH_AES_256_CBC_SHA',
    b'\x00\x35': 'TLS_RSA_WITH_AES_256_CBC_SHA',
    b'\x00\x34': 'TLS_DH_anon_WITH_AES_128_CBC_SHA',
    b'\x00\x33': 'TLS_DHE_RSA_WITH_AES_128_CBC_SHA',
    b'\x00\x32': 'TLS_DHE_DSS_WITH_AES_128_CBC_SHA',
    b'\x00\x31': 'TLS_DH_RSA_WITH_AES_128_CBC_SHA',
    b'\x00\x30': 'TLS_DH_DSS_WITH_AES_128_CBC_SHA',
    b'\x00\x2F': 'TLS_RSA_WITH_AES_128_CBC_SHA',
    b'\x00\x07': 'TLS_RSA_WITH_IDEA_CBC_SHA',
    b'\x00\x25': 'TLS_KRB5_WITH_IDEA_CBC_MD5',
    b'\x00\x21': 'TLS_KRB5_WITH_IDEA_CBC_SHA'

}

cipher_suite_parts = {

    "CryptoAlgo": {
        "AES": (AES, 0),
        "3DES": (TripleDES, 0),
        "CHACHA20": (aead.ChaCha20Poly1305, 1),
        "RC4": (ARC4, 0),
        "CAMELLIA": (Camellia, 0),
        "IDEA": (IDEA,0),

        "GCM": (aead.AESGCM, 1),
        "CCM": (aead.AESCCM, 1)

    },

    "Mode": {
        "CBC": (modes.CBC, 0),  # Tuple[0] is the mode and Tuple[1] is 1 if mode uses AEAD, else it is 0
        "CFB": (modes.CFB, 0),
        "CTR": (modes.CTR, 0),
        "GCM": (modes.GCM, 1),
        "CCM": (aead.AESCCM, 1),
        "POLY1305": (aead.ChaCha20Poly1305, 1)
    },

    "KeyLength": {
        "IDEA": 16,
        "AES_128": 16,
        "AES_256": 32,
        "CAMELLIA_128": 16,
        "CAMELLIA_256": 32,
        "CHACHA20": 32,
        "RC4_128": 16,
        "3DES": 24
    },

    "MAC": {
        "SHA256": hashes.SHA256,
        "SHA384": hashes.SHA384,
        "SHA": hashes.SHA1,
        "MD5": hashes.MD5
    },

    "TagLength": {
        "_8": 8
    }
}


def split_cipher_suite(suite_id):
    try:
        suite_string = cipher_suites[suite_id]
    except KeyError:
        logging.error(f"Cipher suite with value '{suite_id.hex()}' not supported")
        return

    added_part = 0
    cipher_suite = {}

    for part in cipher_suite_parts:
        for p in cipher_suite_parts[part]:
            if p in suite_string:
                cipher_suite.update({part: cipher_suite_parts[part][p]})
                added_part = 1
                break
        if not added_part:
            if part == "TagLength":  # if no tag-length is explicitly given, the tag-length is set to 16 bit
                cipher_suite.update({part: 16})
            else:
                cipher_suite.update({part: (None, 0)})
        added_part = 0

    if cipher_suite["CryptoAlgo"] == (AES,
                                      0):  # if cipher suite uses AES, check if mode is AEAD, so the correct classes from 'cryptography'-module are used
        if "GCM" in suite_string:
            cipher_suite.update({"CryptoAlgo": (aead.AESGCM, 1)})
        elif "CCM" in suite_string:
            cipher_suite.update({"CryptoAlgo": (aead.AESCCM, 1)})

    if cipher_suite["MAC"] == (None, 0):
        cipher_suite["MAC"] = hashes.SHA256

    return cipher_suite



import os
from setuptools import setup, find_packages
from os.path import abspath, dirname, join

__version__ = 0.3
__author__ = "Jannis Borg-Olivier, Lars Morkovsky, Daniel Baier"

# Fetches the content from README.md
# This will be used for the "long_description" field.
README_MD = open(join(dirname(abspath(__file__)), "README.md")).read()

# here - where we are.
here = os.path.abspath(os.path.dirname(__file__))

# read the package requirements for install_requires
with open(os.path.join(here, 'requirements.txt'), 'r') as f:
    requirements = f.readlines()




setup(
    # pip install tlexport
    name="TLExport",
    version=__version__,

    # The description that will be shown on PyPI.
    description="TLExport (TLE) is a tool for decrypting TLS-Traffic and exporting the traffic into unencrypted TCP/UDP traffic",

    # The content that will be shown on your project page.
    # In this case, we're displaying whatever is there in our README.md file
    long_description=README_MD,

    # Now, we'll tell PyPI what language our README file is in.
    long_description_content_type="text/markdown",


    url="https://github.com/fkie-cad/TLExport/",

    author_name=__author__,
    author_email="daniel.baier@fkie.fraunhofer.de",
    license='GPL v3',

     # include other files
     #package_data={
     #   '': [ os.path.join(here, 'TBD') # the frida agent to do the actual hooking
     #    ],  
     # },



    include_package_data=True,
    python_requires='>=3.10',
    packages=find_packages(),
    install_requires=requirements,


    classifiers=[
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Natural Language :: English",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: JavaScript",
        "Topic :: Security"
    ],

    # Keywords are tags that identify your project and help searching for it
    # This field is OPTIONAL
    keywords=["tls", "decryption", "network forensik", "pcap", "pcapng"],

    entry_points={
            'console_scripts': [
            'tlexport=tlexport.main:run',
        ],
    },
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import dpkt
import argparse
import logging

from .packet import Packet
from . import keylog_reader
from .dpkt_dsb import Reader
from .session import Session
from .checksums import calculate_checksum_tcp
from .log import set_logger

server_ports = [443]
keylog = []
sessions = []


def arg_parser_init():
    parser = argparse.ArgumentParser(description="TLExport - GENERATING DECRYPTED TLS PCAPS")
    parser.add_argument("-p", "--serverports", help="additional ports to test for TLS-Connections", nargs="+",
                        default=[443])
    parser.add_argument("-i", "--infile", help="path of input file",
                        default="in.pcapng")
    parser.add_argument("-o", "--outfile", help="path of output file", default="out.pcapng")
    parser.add_argument("-s", "--sslkeylog", help="path to sslkeylogfile")
    # default False due to checksum offloading producing wrong checksums in Packet Capture
    parser.add_argument("-c", "--checksumTest", help="enable for checking tcp Checksums",
                        action=argparse.BooleanOptionalAction, default=False)
    parser.add_argument("-l", "--pcaplegacy", help="enable flag if infile is in the pcap format",
                        action=argparse.BooleanOptionalAction)
    parser.add_argument("-m", "--mapports",
                        help="map TLS-Ports to specific output ports, use this option when using more than one "
                             "serverport <serverport:outputport>",
                        nargs="+",
                        default=["443:8080"])
    parser.add_argument("-d", "--debug", nargs='?', const="INFO", default="ERROR",
                        help="Set the logging level (DEBUG, INFO, WARNING, ERROR)")

    parser.add_argument("-f", "--filter",
                        help="filter log messages by file, add files you want to filter", nargs="+")

    return parser.parse_args()


def get_port_map(parser):
    port_map = {}
    i: str
    for i in parser.mapports:
        split = i.split(":")
        server_port = int(split[0])
        output_port = int(split[1])
        port_map[server_port] = output_port

    return port_map


def handle_packet(packet: Packet, args, keylog, sessions: list[Session], portmap):
    for session in sessions:
        if session.matches_session(packet):
            session.handle_packet(packet)
            return

    if packet.dport in server_ports or packet.sport in server_ports:
        sessions.append(Session(packet, server_ports, keylog, portmap))


def run():

    args = arg_parser_init()
    portmap = get_port_map(args)

    set_logger(args)

    logging.debug(f"Arguments: {args}")
    logging.info(f"Mapping Ports: {portmap}")

    server_ports.extend([int(x) for x in args.serverports])
    


    if args.sslkeylog is not None:
        print(f"[*] Using keys from SSLKEYLOG: {args.sslkeylog}")
        keylog.extend(keylog_reader.read_keylog_from_file(args.sslkeylog))
    else:
        print("[*] Using keys from DSB")

    file = open(args.infile, "rb")

    if args.pcaplegacy:
        pcap_reader = dpkt.pcap.Reader(file)
    else:
        pcap_reader = Reader(file)

    
    print(f"[*] Checking for TLS Traffic on these ports: {server_ports}")

    for ts, buf in pcap_reader:
        packet = Packet(buf, ts)

        if ts == -1:
            keylog.extend(keylog_reader.get_keys_from_string(buf.decode('ascii')))
            continue

        if len(packet.tls_data) == 0:
            continue

        if not args.checksumTest:
            checksum_test = True
        else:
            checksum_test = calculate_checksum_tcp(packet)

        if not checksum_test:
            logging.info("")
            logging.info(f"bad checksum discarded Packet {packet.get_params()}")
            logging.info("")
        if packet.tls_packet and checksum_test:
            handle_packet(packet, args, keylog, sessions, portmap)

    file.close()
    all_decrypted_sessions = []
    for session in sessions:
        all_decrypted_sessions.extend(session.decrypt())

    file = open(args.outfile, "wb")

    writer = dpkt.pcapng.Writer(file)

    for buf, ts in all_decrypted_sessions:
        writer.writepkt(bytes(buf), ts)

    file.close()
    print(f"[*] written decrypted PCAP to {args.outfile}")
    print("[*] Thx for using TLExport. Have a nice day!")


if __name__ == "__main__":
    run()



import logging
# suppress scapy warning message when importing the module
logging.getLogger("scapy.runtime").setLevel(logging.ERROR)
from scapy.packet import Raw
from scapy.layers.l2 import Ether
from scapy.layers.inet import IP, TCP
from tlexport.tlsrecord import TlsRecord
from math import floor


class OutputBuilder:
    def __init__(self, decrypted_records, server_ip, client_ip, server_port, client_port, server_mac_addr,
                 client_mac_addr, portmap) -> None:
        self.decrypted_records = decrypted_records
        self.server_ip = '.'.join(f'{c}' for c in server_ip)
        self.client_ip = '.'.join(f'{c}' for c in client_ip)
        self.server_port = server_port
        self.client_port = client_port
        self.default_port = 8080
        self.server_mac_addr = server_mac_addr
        self.client_mac_addr = client_mac_addr
        self.out = []

        self.server_seq = 1
        self.client_seq = 1

        if self.server_port in portmap.keys():
            self.server_port = portmap[self.server_port]
        else:
            self.server_port = self.default_port

    def build(self):
        self.no_application_records = True

        for record in self.decrypted_records:
            if record is not None:
                self.no_application_records = False

        if self.no_application_records:
            return []

        self.decrypted_records: list[TlsRecord]
        record: (str, TlsRecord, bool)

        self.conn_reset = True

        for record in self.decrypted_records:
            if record is None:
                self.conn_reset = True
                self.server_seq = 1
                self.client_seq = 1
                continue

            if self.conn_reset:
                self.ts_zero = record[1].metadata[0].timestamp
                self.build_ack_handshake()
                self.conn_reset = False

            decrypted = record[0]
            ts = []
            for packet in record[1].metadata:
                ts.append(packet.timestamp)

            if record[2]:
                self.build_server_packet(decrypted, ts)
            else:
                self.build_client_packet(decrypted, ts)
        return self.out

    # due to the size difference between plaintext and ciphertext new Syn/Ack values starting at 0 are used.
    # timestamps for the TCP Handshake are equal to the timestamp of the first TLS-Record
    def build_ack_handshake(self):
        syn = Ether(src=self.client_mac_addr, dst=self.server_mac_addr) / IP(src=self.client_ip,
                                                                             dst=self.server_ip) / TCP(
            dport=self.server_port, sport=self.client_port, flags='S', seq=0, ack=0)
        syn_ack = Ether(src=self.server_mac_addr, dst=self.client_mac_addr) / IP(src=self.server_ip,
                                                                                 dst=self.client_ip) / TCP(
            dport=self.client_port, sport=self.server_port, flags='SA', seq=0, ack=1)
        ack = Ether(src=self.client_mac_addr, dst=self.server_mac_addr) / IP(src=self.client_ip,
                                                                             dst=self.server_ip) / TCP(
            dport=self.server_port, sport=self.client_port, flags='A', seq=1, ack=1)

        self.out.extend([(syn, self.ts_zero), (syn_ack, self.ts_zero), (ack, self.ts_zero)])

    def build_server_packet(self, decrypted, ts):
        record_len = len(decrypted)
        packet_count = len(ts)
        part_len = floor(record_len / packet_count)
        parts = []
        last_len = 0
        for i in range(0, packet_count - 1):
            parts.append(decrypted[i * part_len: i * part_len + part_len])
            last_len = i * part_len + part_len

        if last_len < record_len:
            parts.append(decrypted[last_len:])
        for i in range(0, len(parts)):
            packet = Ether(src=self.server_mac_addr, dst=self.client_mac_addr) / IP(src=self.server_ip,
                                                                                    dst=self.client_ip) / TCP(
                dport=self.client_port, sport=self.server_port, flags='PA', seq=self.server_seq,
                ack=self.client_seq) / Raw(parts[i])
            self.server_seq += len(parts[i])
            packet_ack = Ether(src=self.client_mac_addr, dst=self.server_mac_addr) / IP(src=self.client_ip,
                                                                                        dst=self.server_ip) / TCP(
                dport=self.server_port, sport=self.client_port, flags='A', seq=self.client_seq, ack=self.server_seq)
            self.out.append((packet, ts[i]))
            self.out.append((packet_ack, ts[i]))

    def build_client_packet(self, decrypted, ts):
        record_len = len(decrypted)
        packet_count = len(ts)
        part_len = floor(record_len / packet_count)
        parts = []
        last_len = 0
        for i in range(0, packet_count - 1):
            parts.append(decrypted[i * part_len: i * part_len + part_len])
            last_len = i * part_len + part_len

        if last_len < record_len:
            parts.append(decrypted[last_len:])

        for i in range(0, len(parts)):
            packet = Ether(src=self.client_mac_addr, dst=self.server_mac_addr) / IP(src=self.client_ip,
                                                                                    dst=self.server_ip) / TCP(
                dport=self.server_port, sport=self.client_port, flags='PA', seq=self.client_seq,
                ack=self.server_seq) / Raw(parts[i])
            self.client_seq += len(parts[i])
            packet_ack = Ether(src=self.server_mac_addr, dst=self.client_mac_addr) / IP(src=self.server_ip,
                                                                                        dst=self.client_ip) / TCP(
                dport=self.client_port, sport=self.server_port, flags='A', seq=self.server_seq, ack=self.client_seq)
            self.out.append((packet, ts[i]))
            self.out.append((packet_ack, ts[i]))
